const fs = require('fs-extra')
const path = require('path')
const ora = require('ora')
const kebabCase = require('lodash.kebabcase')
const { print, run, prompt } = require('@toptal/davinci-cli-shared')

const skeletonLibMainPath = require.resolve('@toptal/davinci-skeleton')

const NODE_MODULES = /skeleton\/node_modules/g

const checkPrerequisites = async (currentRunningDir, appName) => {
  const spinner = ora('Checking prerequisites...').start()

  // look for env resolvers
  try {
    await run('direnv', ['version'])
  } catch (err) {
    spinner.warn()
    print.yellow(
      "warning: You don't have `direnv` for resolving env variables! Expect other approach is used."
    )
  }

  // check if app already exists
  const appPath = path.join(currentRunningDir, appName)
  const doesAppExists = await fs.exists(appPath)

  if (doesAppExists) {
    spinner.fail()
    throw new Error(`Application with name ${appName} already exists!`)
  }

  spinner.succeed('Prerequisites checked!')
}

const aquirePrerequisites = async () => {
  const { NPM_TOKEN: envNpmToken } = process.env

  if (envNpmToken) {
    const shouldUse = await prompt.confirm(
      'envNpmToken',
      'NPM_TOKEN detected! Should we use it for accessing @toptal?'
    )

    if (shouldUse) {
      return { npmToken: envNpmToken }
    }
  }

  const newNpmToken = await prompt.password(
    'npmToken',
    'Enter npm token:',
    newNpmToken =>
      (newNpmToken && newNpmToken.length > 0) ||
      'NPM token is not valid. Go to https://www.npmjs.com/settings/talbot/tokens.'
  )

  return {
    npmToken: newNpmToken
  }
}

const copyApplicationSkeleton = async (dest, appName) => {
  const appDir = path.join(skeletonLibMainPath, '../..')
  const destDir = path.join(dest, appName)

  const spinner = ora('Copying skeleton...').start()

  const ignoreSkeletonFiles = src => !src.match(NODE_MODULES)

  try {
    await fs.copy(appDir, destDir, { filter: ignoreSkeletonFiles })
    spinner.succeed('Skeleton copied!')
  } catch (err) {
    spinner.fail()
    throw err
  }
}

const modifyPackageJson = async (dest, appName) => {
  const packageJsonFile = path.join(dest, appName, './package.json')

  const spinner = ora('Modifiying package.json...').start()

  try {
    const configData = fs.readFileSync(packageJsonFile)
    const config = JSON.parse(configData)

    config.name = `@toptal/${appName}`

    const modifiedConfigData = JSON.stringify(config, null, 2).concat('\n')

    fs.writeFileSync(packageJsonFile, modifiedConfigData)

    spinner.succeed('package.json modified!')
  } catch (err) {
    spinner.fail()
    throw err
  }
}

const allowDirEnv = async appName => {
  try {
    await run('direnv', ['allow', appName])
  } catch (err) {
    // suppress errors because it is optional
  }
}

const createEnvrcFile = async (dest, appName, npmToken) => {
  const envrcFileFile = path.join(dest, appName, './.envrc')

  const spinner = ora('Creating .envrc file ...').start()

  try {
    await fs.writeFile(envrcFileFile, `export NPM_TOKEN=${npmToken}`)
    await allowDirEnv(appName)

    spinner.succeed('.envrc created!')
  } catch (err) {
    spinner.fail()
    throw err
  }
}

const installDependencies = async (dest, appName, npmToken) => {
  const appDir = path.join(dest, appName)

  const spinner = ora('Installing dependencies... ').start()

  try {
    await run('yarn', [], { cwd: appDir, env: { NPM_TOKEN: npmToken } })
    spinner.succeed('Dependencies installed!')
  } catch (err) {
    spinner.fail()
    throw new Error(err)
  }
}

const createCommand = async appName => {
  print.green('Creating', appName, '...')

  const currentRunningDir = process.cwd()

  try {
    // check if all apps are installed, directory doesn't exists
    await checkPrerequisites(currentRunningDir, appName)

    // aquire any data, metadata needed for whole process
    const { npmToken } = await aquirePrerequisites()

    // copy skeleton to destination directory
    await copyApplicationSkeleton(currentRunningDir, appName)

    // change package.json
    await modifyPackageJson(currentRunningDir, appName)

    // create .envrc file
    await createEnvrcFile(currentRunningDir, appName, npmToken)

    // run yarn
    await installDependencies(currentRunningDir, appName, npmToken)

    print.green('Application created.')
  } catch (err) {
    print.red(err)
  }
}

const newCommandCreator = {
  command: 'new <appName>',
  description: 'Create new application',
  action: async ({ appName }) => {
    const formattedAppName = kebabCase(appName)

    return createCommand(formattedAppName)
  }
}

module.exports = newCommandCreator
